@@ -1,11 +1,17 @@
 #include <stdio.h>
 #include <stdlib.h>
+#include <math.h>
 #include "abpoa_align.h"
 #include "abpoa_seq.h"
 #include "simd_abpoa_align.h"
 #include "kdq.h"
 
-static const char LogTable256[256] = {
+extern char ab_char256_table[256];
+char ab_LogTable65536[65536];
+char ab_bit_table16[65536];
+
+#define NAT_E 2.718281828459045
+static const char ab_LogTable256[256] = {
 #define LT(n) n, n, n, n, n, n, n, n, n, n, n, n, n, n, n, n
     -1, 0, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3,
     LT(4), LT(5), LT(5), LT(6), LT(6), LT(6), LT(6),
@@ -15,29 +21,28 @@ static const char LogTable256[256] = {
 static inline int ilog2_32(uint32_t v)
 {
     uint32_t t, tt;
-    if ((tt = v>>16)) return (t = tt>>8) ? 24 + LogTable256[t] : 16 + LogTable256[tt];
-    return (t = v>>8) ? 8 + LogTable256[t] : LogTable256[v];
+    if ((tt = v>>16)) return (t = tt>>8) ? 24 + ab_LogTable256[t] : 16 + ab_LogTable256[tt];
+    return (t = v>>8) ? 8 + ab_LogTable256[t] : ab_LogTable256[v];
 }
 
-void set_65536_table(abpoa_para_t *abpt) {
+void set_65536_table(void) {
     int i;
     for (i = 0; i < 65536; ++i) {
-        abpt->LogTable65536[i] = ilog2_32(i);
+        ab_LogTable65536[i] = ilog2_32(i);
     }
 }
 
-void set_bit_table16(abpoa_para_t *abpt) {
-    int i; abpt->bit_table16[0] = 0;
-    for (i = 0; i != 65536; ++i) abpt->bit_table16[i] = (i&1) + abpt->bit_table16[i>>1];
+void set_bit_table16(void) {
+    int i; ab_bit_table16[0] = 0;
+    for (i = 0; i != 65536; ++i) ab_bit_table16[i] = (i&1) + ab_bit_table16[i>>1];
 }
 
 #define get_bit_cnt4(table, b) (table[(b)&0xffff] + table[(b)>>16&0xffff] + table[(b)>>32&0xffff] + table[(b)>>48&0xffff])
 
-static inline int ilog2_64(abpoa_para_t *abpt, uint64_t v)
-{
+static inline int ilog2_64(uint64_t v) {
     uint64_t t, tt;
-    if ((tt = v >> 32)) return (t = tt >> 16) ? 48 + abpt->LogTable65536[t] : 32 + abpt->LogTable65536[tt];
-    return (t = v>>16) ? 16 + abpt->LogTable65536[t] : abpt->LogTable65536[v];
+    if ((tt = v >> 32)) return (t = tt >> 16) ? 48 + ab_LogTable65536[t] : 32 + ab_LogTable65536[tt];
+    return (t = v>>16) ? 16 + ab_LogTable65536[t] : ab_LogTable65536[v];
 }
 
 KDQ_INIT(int)
@@ -143,10 +148,11 @@ void abpoa_free(abpoa_t *ab) {
 }
 
 void abpoa_BFS_set_node_index(abpoa_graph_t *abg, int src_id, int sink_id) {
-    int *id, cur_id, i, j, out_id, aligned_id, q_size, new_q_size;
-    int index = 0;
+    int *id, cur_id, out_id, aligned_id;
+    int index = 0, q_size, new_q_size;
 
     int *in_degree = (int*)_err_malloc(abg->node_n * sizeof(int));
+    int i, j;
     for (i = 0; i < abg->node_n; ++i) in_degree[i] = abg->node[i].in_edge_n;
 
     kdq_int_t *q = kdq_init_int();
@@ -332,7 +338,7 @@ void abpoa_set_msa_rank(abpoa_graph_t *abg, int src_id, int sink_id) {
     }
 }
 
-void abpoa_set_row_column_weight(abpoa_graph_t *abg, abpoa_para_t *abpt, int **rc_weight, int **msa_node_id) {
+void abpoa_set_row_column_weight(abpoa_graph_t *abg, int **rc_weight, int **msa_node_id) {
     int i, k, rank, aligned_id;
     uint64_t b;
     for (i = 2; i < abg->node_n; ++i) {
@@ -345,14 +351,14 @@ void abpoa_set_row_column_weight(abpoa_graph_t *abg, abpoa_para_t *abpt, int **r
         // assign seq
         for (k = 0; k < abg->node[i].read_ids_n; ++k) {
             b = abg->node[i].read_ids[k];
-            rc_weight[rank-1][abg->node[i].base] += get_bit_cnt4(abpt->bit_table16, b);
+            rc_weight[rank-1][abg->node[i].base] += get_bit_cnt4(ab_bit_table16, b);
         }
         rc_weight[rank-1][4] -= rc_weight[rank-1][abg->node[i].base];
         msa_node_id[rank-1][abg->node[i].base] = i;
     }
 }
 
-void abpoa_set_row_column_ids_weight(abpoa_graph_t *abg, abpoa_para_t *abpt, uint64_t ***read_ids, int **rc_weight, int **msa_node_id, int msa_l, int n_seq, int read_ids_n) {
+void abpoa_set_row_column_ids_weight(abpoa_graph_t *abg, uint64_t ***read_ids, int **rc_weight, int **msa_node_id, int msa_l, int n_seq, int read_ids_n) {
     int i, j, k, rank, aligned_id;
     uint64_t b, one = 1, *whole_read_ids = (uint64_t*)_err_calloc(read_ids_n, sizeof(uint64_t));
     for (i = 0; i < n_seq; ++i) {
@@ -374,7 +380,7 @@ void abpoa_set_row_column_ids_weight(abpoa_graph_t *abg, abpoa_para_t *abpt, uin
         // assign seq
         for (k = 0; k < abg->node[i].read_ids_n; ++k) {
             b = abg->node[i].read_ids[k];
-            rc_weight[rank-1][abg->node[i].base] += get_bit_cnt4(abpt->bit_table16, b);
+            rc_weight[rank-1][abg->node[i].base] += get_bit_cnt4(ab_bit_table16, b);
             read_ids[rank-1][abg->node[i].base][k] = b;
             read_ids[rank-1][4][k] ^= b;
         }
@@ -384,7 +390,7 @@ void abpoa_set_row_column_ids_weight(abpoa_graph_t *abg, abpoa_para_t *abpt, uin
     free(whole_read_ids);
 }
 
-void abpoa_heaviest_column_multip_consensus(abpoa_para_t *abpt, uint64_t ***read_ids, int **cluster_ids, int *cluster_ids_n, int multip, int read_ids_n, int msa_l, FILE *out_fp, uint8_t ***_cons_seq, int **_cons_l, int *_cons_n) {
+void abpoa_heaviest_column_multip_consensus(uint64_t ***read_ids, int **cluster_ids, int *cluster_ids_n, int multip, int read_ids_n, int msa_l, FILE *out_fp, uint8_t ***_cons_seq, int **_cons_l, int *_cons_n) {
     int i, j, k, m, w, cnt, max_w, max_base, gap_w;
     uint64_t *read_ids_mask = (uint64_t*)_err_malloc(read_ids_n * sizeof(uint64_t)), one=1, b;
     uint8_t *cons_seq = (uint8_t*)_err_malloc(sizeof(uint8_t) * msa_l); int cons_l;
@@ -410,7 +416,7 @@ void abpoa_heaviest_column_multip_consensus(abpoa_para_t *abpt, uint64_t ***read
                 w = 0;
                 for (m = 0; m < read_ids_n; ++m) {
                     b = read_ids[j][k][m] & read_ids_mask[m];
-                    cnt = get_bit_cnt4(abpt->bit_table16, b);
+                    cnt = get_bit_cnt4(ab_bit_table16, b);
                     w += cnt, gap_w -= cnt;
                 }
                 if (w > max_w) {
@@ -422,7 +428,7 @@ void abpoa_heaviest_column_multip_consensus(abpoa_para_t *abpt, uint64_t ***read
             }
         }
         if (out_fp) {
-            for (j = 0; j < cons_l; ++j) fprintf(out_fp, "%c", "ACGT"[cons_seq[j]]); fprintf(out_fp, "\n");
+            for (j = 0; j < cons_l; ++j) fprintf(out_fp, "%c", ab_char256_table[cons_seq[j]]); fprintf(out_fp, "\n");
         }
         if (_cons_n) {
             (*_cons_l)[i] = cons_l;
@@ -433,15 +439,28 @@ void abpoa_heaviest_column_multip_consensus(abpoa_para_t *abpt, uint64_t ***read
     free(read_ids_mask); free(cons_seq);
 }
 
-int output_consensus(abpoa_graph_t *abg, int src_id, int sink_id, FILE *out_fp) {
+int output_consensus(int out_fq, abpoa_graph_t *abg, int src_id, int sink_id, int **cons_cov, int n_seq, FILE *out_fp) {
     int cons_l = 0;
-    fprintf(out_fp, ">Consensus_sequence\n");
+    if (out_fq) fprintf(out_fp, "@Consensus_sequence\n");
+    else fprintf(out_fp, ">Consensus_sequence\n");
     int id = abg->node[src_id].max_out_id;
     while (id != sink_id) {
-        fprintf(out_fp, "%c", "ACGTN"[abg->node[id].base]);
+        fprintf(out_fp, "%c", ab_char256_table[abg->node[id].base]);
         id = abg->node[id].max_out_id;
         cons_l++;
     } fprintf(out_fp, "\n");
+    if (out_fq) {
+        fprintf(out_fp, "+Consensus_sequence\n");
+        if (cons_cov == NULL) err_fatal_simple("No coverage information available.");
+        int i, phred; double x, p;
+        for (i = 0; i < cons_l; ++i) {
+            // max q-score: 33+60=93, min score: 33+0=33
+            x = 13.8 * (1.25 * cons_cov[0][i] / n_seq - 0.25);
+            p = 1 - 1.0 / (1.0 + pow(NAT_E, -1 * x));
+            phred = 33 + (int)(-10 * log10(p) + 0.499);
+            fprintf(out_fp, "%c", phred);
+        } fprintf(out_fp, "\n");
+    }
     return cons_l;
 }
 
@@ -483,8 +502,8 @@ void abpoa_generate_consensus_cov(abpoa_graph_t *abg, int src_id, int sink_id, i
     }
 }
 
-void abpoa_traverse_min_flow(abpoa_graph_t *abg,  int src_id, int sink_id, int *out_degree) {
-    int *id, i, cur_id, in_id, out_id, max_id, max_w, max_out_i, min_w;
+void abpoa_traverse_min_flow(abpoa_graph_t *abg, int src_id, int sink_id, int *out_degree) {
+    int *id, i, cur_id, in_id, out_id, max_id; int max_w, max_out_i, min_w;
     kdq_int_t *q = kdq_init_int(); kdq_push_int(q, sink_id);
     int *max_out_weight = (int*)_err_malloc(abg->node_n * sizeof(int));
     // reverse Breadth-First-Search
@@ -527,8 +546,8 @@ MF_CONS:
 // heaviest_bundling
 // 1. argmax{cur->weight}
 // 2. argmax{out_node->weight}
-void abpoa_heaviest_bundling(abpoa_graph_t *abg,  int src_id, int sink_id, int *out_degree, int ***cons_cov) {
-    int *id, i, cur_id, in_id, out_id, max_id, max_w, out_w;
+void abpoa_heaviest_bundling(abpoa_graph_t *abg, int src_id, int sink_id, int *out_degree, int ***cons_cov) {
+    int *id, i, cur_id, in_id, out_id, max_id; int max_w, out_w;
 
     int *score = (int*)_err_malloc(abg->node_n * sizeof(int));
     kdq_int_t *q = kdq_init_int();
@@ -582,9 +601,13 @@ HB_CONS:
     free(score);
 }
 
-void abpoa_heaviest_column_consensus(abpoa_graph_t *abg, int **rc_weight, int **msa_node_id, int src_id, int sink_id, int msa_l, int n_seq) {
+void abpoa_heaviest_column_consensus(abpoa_graph_t *abg, int **rc_weight, int **msa_node_id, int src_id, int sink_id, int msa_l, int n_seq, int ***cons_cov) {
     int i, j, w, max_base, max_w, gap_w;
-    int last_id = src_id, cur_id;
+    int last_id = src_id, cur_id, cons_i = 0;
+    if (cons_cov != NULL) {
+        *cons_cov = (int**)_err_malloc(sizeof(int*));
+        (*cons_cov)[0] = (int*)_err_malloc(sizeof(int) * abg->node_n);
+    }
     for (i = 0; i < msa_l; ++i) {
         max_w = 0, max_base = 5, gap_w = n_seq;
         for (j = 0; j < 4; ++j) {
@@ -598,12 +621,14 @@ void abpoa_heaviest_column_consensus(abpoa_graph_t *abg, int **rc_weight, int **
             cur_id = msa_node_id[i][max_base];
             abg->node[last_id].max_out_id = cur_id;
             last_id = cur_id;
+            if (cons_cov != NULL) 
+                (*cons_cov)[0][cons_i++] = max_w;
         }
     }
     abg->node[last_id].max_out_id = sink_id;
 }
 
-void abpoa_heaviest_column(abpoa_graph_t *abg, abpoa_para_t *abpt, int src_id, int sink_id, int n_seq) {
+void abpoa_heaviest_column(abpoa_graph_t *abg, int src_id, int sink_id, int n_seq, int ***cons_cov) {
     abpoa_set_msa_rank(abg, src_id, sink_id);
 
     int i, msa_l = abg->node_id_to_msa_rank[sink_id] - 1;
@@ -614,15 +639,15 @@ void abpoa_heaviest_column(abpoa_graph_t *abg, abpoa_para_t *abpt, int src_id, i
         rc_weight[i][4] = n_seq;
         msa_node_id[i] = (int*)_err_calloc(5, sizeof(int));
     } 
-    abpoa_set_row_column_weight(abg, abpt, rc_weight, msa_node_id);
+    abpoa_set_row_column_weight(abg, rc_weight, msa_node_id);
 
-    abpoa_heaviest_column_consensus(abg, rc_weight, msa_node_id, src_id, sink_id, msa_l, n_seq);
+    abpoa_heaviest_column_consensus(abg, rc_weight, msa_node_id, src_id, sink_id, msa_l, n_seq, cons_cov);
     for (i = 0; i < msa_l; ++i) { 
         free(rc_weight[i]); free(msa_node_id[i]);
     } free(rc_weight); free(msa_node_id);
 }
 
-void add_het_read_ids(abpoa_para_t *abpt, int *init, uint64_t **het_read_ids, uint8_t **het_cons_base, uint64_t **read_ids, int het_n, int *multip_i, int read_ids_n) {
+void add_het_read_ids(int *init, uint64_t **het_read_ids, uint8_t **het_cons_base, uint64_t **read_ids, int het_n, int *multip_i, int read_ids_n) {
     int i, j;
     if (*init) {
         for (i = 0; i < 2; ++i) {
@@ -640,7 +665,7 @@ void add_het_read_ids(abpoa_para_t *abpt, int *init, uint64_t **het_read_ids, ui
             ovlp_n = 0; 
             for (j = 0; j < read_ids_n; ++j) {
                 b = het_read_ids[i][j] & ids[j];
-                ovlp_n += get_bit_cnt4(abpt->bit_table16, b);
+                ovlp_n += get_bit_cnt4(ab_bit_table16, b);
             }
             if (ovlp_n > max_ovlp_n) {
                 max_ovlp_n = ovlp_n;
@@ -691,7 +716,7 @@ int set_clu_read_ids(int **clu_read_ids, int *clu_read_ids_n, int n_seq, double
     return clu_n;
 }
 
-int abpoa_diploid_ids(abpoa_para_t *abpt, uint64_t ***read_ids, int **rc_weight, int msa_l, int n_seq, double min_freq, int read_ids_n, int **clu_read_ids, int *clu_read_ids_n) {
+int abpoa_diploid_ids(uint64_t ***read_ids, int **rc_weight, int msa_l, int n_seq, double min_freq, int read_ids_n, int **clu_read_ids, int *clu_read_ids_n) {
     int i, j;
     uint64_t **het_read_ids = (uint64_t**)_err_malloc(sizeof(uint64_t*) * 2 );
     for (i = 0; i < 2; ++i) het_read_ids[i] = (uint64_t*)_err_calloc(read_ids_n, sizeof(uint64_t));
@@ -720,7 +745,7 @@ int abpoa_diploid_ids(abpoa_para_t *abpt, uint64_t ***read_ids, int **rc_weight,
             // iteratively update read_ids and cons-base for each cluster
             //     read_ids |=
             //     cons_base1[i++] = ; cons_base2[i++] = ;
-            add_het_read_ids(abpt, &init, het_read_ids, het_cons_base, read_ids[i], het_n, multip_i, read_ids_n);
+            add_het_read_ids(&init, het_read_ids, het_cons_base, read_ids[i], het_n, multip_i, read_ids_n);
             het_pos[het_n++] = i;
         }
     }
@@ -738,7 +763,7 @@ int abpoa_diploid_ids(abpoa_para_t *abpt, uint64_t ***read_ids, int **rc_weight,
     return clu_n;
 }
 
-void abpoa_diploid_heaviest_column(abpoa_graph_t *abg, abpoa_para_t *abpt, int src_id, int sink_id, int n_seq, double min_freq, FILE *out_fp, uint8_t ***cons_seq, int **cons_l, int *cons_n) {
+void abpoa_diploid_heaviest_column(abpoa_graph_t *abg, int src_id, int sink_id, int n_seq, double min_freq, FILE *out_fp, uint8_t ***cons_seq, int **cons_l, int *cons_n) {
     abpoa_set_msa_rank(abg, src_id, sink_id);
     int i, j, msa_l = abg->node_id_to_msa_rank[sink_id] - 1;
     int read_ids_n = (n_seq-1)/64+1;
@@ -754,19 +779,19 @@ void abpoa_diploid_heaviest_column(abpoa_graph_t *abg, abpoa_para_t *abpt, int s
         rc_weight[i][4] = n_seq;
         msa_node_id[i] = (int*)_err_calloc(5, sizeof(int)); // ACGT
     } 
-    abpoa_set_row_column_ids_weight(abg, abpt, read_ids, rc_weight, msa_node_id, msa_l, n_seq, read_ids_n);
+    abpoa_set_row_column_ids_weight(abg, read_ids, rc_weight, msa_node_id, msa_l, n_seq, read_ids_n);
 
     int **clu_read_ids = (int**)_err_malloc(sizeof(int*) * 2), *clu_read_ids_n;
     for (i = 0; i < 2; ++i) clu_read_ids[i] = (int*)_err_malloc(sizeof(int) * n_seq);
     clu_read_ids_n = (int*)_err_calloc(2, sizeof(int));
-    int clu_n = abpoa_diploid_ids(abpt, read_ids, rc_weight, msa_l, n_seq, min_freq, read_ids_n, clu_read_ids, clu_read_ids_n);
+    int clu_n = abpoa_diploid_ids(read_ids, rc_weight, msa_l, n_seq, min_freq, read_ids_n, clu_read_ids, clu_read_ids_n);
     if (clu_n == 1) {
-        abpoa_heaviest_column_consensus(abg, rc_weight, msa_node_id, src_id, sink_id, msa_l, n_seq);
-        if (out_fp) output_consensus(abg, src_id, sink_id, out_fp);
+        abpoa_heaviest_column_consensus(abg, rc_weight, msa_node_id, src_id, sink_id, msa_l, n_seq, NULL);
+        if (out_fp) output_consensus(0, abg, src_id, sink_id, NULL, n_seq, out_fp);
         if (cons_n) {
             *cons_n = 1; abpoa_store_consensus(abg, src_id, sink_id, cons_seq, cons_l);
         }
-    } else abpoa_heaviest_column_multip_consensus(abpt, read_ids, clu_read_ids, clu_read_ids_n, clu_n, read_ids_n, msa_l, out_fp, cons_seq, cons_l, cons_n);
+    } else abpoa_heaviest_column_multip_consensus(read_ids, clu_read_ids, clu_read_ids_n, clu_n, read_ids_n, msa_l, out_fp, cons_seq, cons_l, cons_n);
 
     for (i = 0; i < msa_l; ++i) {
         for (j = 0; j < 5; ++j) {
@@ -782,29 +807,39 @@ int abpoa_generate_consensus(abpoa_t *ab, abpoa_para_t *abpt, FILE *out_fp, uint
     abpoa_graph_t *abg = ab->abg;
     if (abg->node_n <= 2) return 0;
     int i, _cons_l = 0, *out_degree = (int*)_err_malloc(abg->node_n * sizeof(int));
+    int **_cons_cov;
+    if (abpt->out_fq == 0 && cons_cov == NULL) _cons_cov = NULL;
     int n_seq = ab->abs->n_seq;
     for (i = 0; i < abg->node_n; ++i) {
         out_degree[i] = abg->node[i].out_edge_n;
     }
 
     if (abpt->is_diploid) {
-        abpoa_diploid_heaviest_column(abg, abpt, ABPOA_SRC_NODE_ID, ABPOA_SINK_NODE_ID, n_seq, abpt->min_freq, out_fp, cons_seq, cons_l, cons_n);
+        abpoa_diploid_heaviest_column(abg, ABPOA_SRC_NODE_ID, ABPOA_SINK_NODE_ID, n_seq, abpt->min_freq, out_fp, cons_seq, cons_l, cons_n);
     } else {
-        if (abpt->cons_agrm == ABPOA_HB) abpoa_heaviest_bundling(abg, ABPOA_SRC_NODE_ID, ABPOA_SINK_NODE_ID, out_degree, cons_cov);
-        else if (abpt->cons_agrm == ABPOA_HC) abpoa_heaviest_column(abg, abpt, ABPOA_SRC_NODE_ID, ABPOA_SINK_NODE_ID, n_seq);
-        else if (abpt->cons_agrm == ABPOA_MF) abpoa_traverse_min_flow(abg, ABPOA_SRC_NODE_ID, ABPOA_SINK_NODE_ID, out_degree); 
+        if (abpt->cons_agrm == ABPOA_HB) abpoa_heaviest_bundling(abg, ABPOA_SRC_NODE_ID, ABPOA_SINK_NODE_ID, out_degree, &_cons_cov);
+        else if (abpt->cons_agrm == ABPOA_HC) abpoa_heaviest_column(abg, ABPOA_SRC_NODE_ID, ABPOA_SINK_NODE_ID, n_seq, &_cons_cov);
         else err_fatal(__func__, "Unknown consensus calling algorithm: %d.", abpt->cons_agrm);
-        if (out_fp) _cons_l = output_consensus(abg, ABPOA_SRC_NODE_ID, ABPOA_SINK_NODE_ID, out_fp);
-        if (cons_n) {
+        if (out_fp) _cons_l = output_consensus(abpt->out_fq, abg, ABPOA_SRC_NODE_ID, ABPOA_SINK_NODE_ID, _cons_cov, n_seq, out_fp);
+        if (cons_n && cons_l && cons_seq) {
             *cons_n = 1; 
             _cons_l = abpoa_store_consensus(abg, ABPOA_SRC_NODE_ID, ABPOA_SINK_NODE_ID, cons_seq, cons_l);
         }
+        if (cons_cov && _cons_l) {
+            *cons_cov = (int**)_err_malloc(sizeof(int*));
+            (*cons_cov)[0] = (int*)_err_malloc(sizeof(int) * _cons_l);
+            for (i = 0; i < _cons_l; ++i) (*cons_cov)[0][i] = _cons_cov[0][i];
+        }
+
+        if (_cons_cov) {
+            free(_cons_cov[0]); free(_cons_cov);
+        }
     } 
     abg->is_called_cons = 1; free(out_degree);
     return _cons_l;
 }
 
-void abpoa_set_msa_seq(abpoa_para_t *abpt, abpoa_node_t node, int rank, uint8_t **msa_seq) {
+void abpoa_set_msa_seq(abpoa_node_t node, int rank, uint8_t **msa_seq) {
     int i, b, read_id; uint8_t base = node.base;
     uint64_t num, tmp;
     b = 0;
@@ -812,7 +847,7 @@ void abpoa_set_msa_seq(abpoa_para_t *abpt, abpoa_node_t node, int rank, uint8_t
         num = node.read_ids[i];
         while (num) {
             tmp = num & -num;
-            read_id = ilog2_64(abpt, tmp);
+            read_id = ilog2_64(tmp);
             // fprintf(stderr, "%d -> %d\n", node.node_id, read_id);
             msa_seq[b+read_id][rank-1] = base;
             num ^= tmp;
@@ -834,7 +869,7 @@ void abpoa_generate_rc_msa(abpoa_t *ab, abpoa_para_t *abpt, FILE *out_fp, uint8_
     for (i = 0; i < n_seq; ++i) {
         _msa_seq[i] = (uint8_t*)_err_malloc(_msa_l * sizeof(uint8_t));
         for (j = 0; j < _msa_l; ++j) 
-            _msa_seq[i][j] = 5;
+            _msa_seq[i][j] = 27;
     }
 
     if (out_fp && abpt->out_msa_header == 0) fprintf(out_fp, ">Multiple_sequence_alignment\n");
@@ -846,7 +881,7 @@ void abpoa_generate_rc_msa(abpoa_t *ab, abpoa_para_t *abpt, FILE *out_fp, uint8_
             rank = MAX_OF_TWO(rank, abpoa_graph_node_id_to_msa_rank(abg, aligned_id));
         }
         // assign seq
-        abpoa_set_msa_seq(abpt, abg->node[i], rank, _msa_seq);
+        abpoa_set_msa_seq(abg->node[i], rank, _msa_seq);
     }
     if (out_fp) {
         for (i = 0; i < n_seq; ++i) {
@@ -854,7 +889,7 @@ void abpoa_generate_rc_msa(abpoa_t *ab, abpoa_para_t *abpt, FILE *out_fp, uint8_
                 if (abs->is_rc[i]) fprintf(out_fp, ">%s_reverse_complement\n", abs->name[i].s);
                 else fprintf(out_fp, ">%s\n", abs->name[i].s);
             }
-            for (j = 0; j < _msa_l; ++j) fprintf(out_fp, "%c", "ACGTN-"[_msa_seq[i][j]]);
+            for (j = 0; j < _msa_l; ++j) fprintf(out_fp, "%c", ab_char256_table[_msa_seq[i][j]]);
             fprintf(out_fp, "\n");
         }
         if (abpt->out_cons) { // RC-MSA for consensus sequence
@@ -871,7 +906,7 @@ void abpoa_generate_rc_msa(abpoa_t *ab, abpoa_para_t *abpt, FILE *out_fp, uint8_
                 // last_rank -> rank : -
                 for (k = last_rank; k < rank; ++k) fprintf(out_fp, "-");
                 // rank : base
-                fprintf(out_fp, "%c", "ACGTN"[abg->node[i].base]);
+                fprintf(out_fp, "%c", ab_char256_table[abg->node[i].base]);
                 last_rank = rank+1;
                 i = abg->node[i].max_out_id;
             }
@@ -916,7 +951,7 @@ void abpoa_generate_gfa(abpoa_t *ab, abpoa_para_t *abpt, FILE *out_fp) {
         } else {
             if (cur_id != ABPOA_SRC_NODE_ID) {
                 // output node
-                fprintf(out_fp, "S\t%d\t%c\n", cur_id-1, "ACGTN"[abg->node[cur_id].base]);
+                fprintf(out_fp, "S\t%d\t%c\n", cur_id-1, ab_char256_table[abg->node[cur_id].base]);
                 // output all links based pre_ids
                 for (i = 0; i < abg->node[cur_id].in_edge_n; ++i) {
                     pre_id = abg->node[cur_id].in_id[i];
@@ -930,7 +965,7 @@ void abpoa_generate_gfa(abpoa_t *ab, abpoa_para_t *abpt, FILE *out_fp) {
                     num = abg->node[cur_id].read_ids[i];
                     while (num) {
                         tmp = num & -num;
-                        read_id = ilog2_64(abpt, tmp);
+                        read_id = ilog2_64(tmp);
                         read_paths[b+read_id][read_path_i[b+read_id]++] = cur_id-1;
                         num ^= tmp;
                     }
@@ -1070,17 +1105,20 @@ ADD_READ_ID:
     return ret;
 }
 
-void abpoa_add_graph_sequence(abpoa_graph_t *abg, uint8_t *seq, int seq_l, int start, int end, uint8_t add_read_id, int read_id, int read_ids_n) {
-    if (seq_l <= 0 || start >= seq_l || end <= start) err_fatal(__func__, "seq_l: %d\tstart: %d\tend: %d.", seq_l, start, end);
-    if (start < 0) start = 0; if (end > seq_l) end = seq_l;
-    int node_id = abpoa_add_graph_node(abg, seq[start]);
-    abpoa_add_graph_edge(abg, ABPOA_SRC_NODE_ID, node_id, 0, 1, add_read_id, read_id, read_ids_n);
-    int i; 
-    for (i = start+1; i < end; ++i) {
-        node_id = abpoa_add_graph_node(abg, seq[i]);
-        abpoa_add_graph_edge(abg, node_id-1, node_id, 0, 1, add_read_id, read_id, read_ids_n);
+void abpoa_add_graph_sequence(abpoa_graph_t *abg, uint8_t *seq, int seq_l, int *qpos_to_node_id, int start, int end, uint8_t add_read_id, int read_id, int read_ids_n) {
+    if (start >= seq_l || end <= start) err_fatal(__func__, "seq_l: %d\tstart: %d\tend: %d.", seq_l, start, end);
+    if (end > seq_l) end = seq_l;
+
+    int i, last_node_id, cur_node_id;
+    last_node_id = ABPOA_SRC_NODE_ID;
+    for (i = start; i < end; ++i) {
+        cur_node_id = abpoa_add_graph_node(abg, seq[i]);
+        if (qpos_to_node_id) qpos_to_node_id[i] = cur_node_id;
+        abpoa_add_graph_edge(abg, last_node_id, cur_node_id, 0, 1, add_read_id, read_id, read_ids_n);
+        last_node_id = cur_node_id;
     }
-    abpoa_add_graph_edge(abg, node_id, ABPOA_SINK_NODE_ID, 0, 1, add_read_id, read_id, read_ids_n);
+
+    abpoa_add_graph_edge(abg, last_node_id, ABPOA_SINK_NODE_ID, 0, 1, add_read_id, read_id, read_ids_n);
     abg->is_called_cons = abg->is_set_msa_rank = abg->is_topological_sorted = 0;
     // abpoa_topological_sort(abg, abpt);
 }
@@ -1098,15 +1136,16 @@ int is_full_upstream_subgraph(abpoa_graph_t *abg, int up_index, int down_index)
 }
 
 int abpoa_upstream_index(abpoa_graph_t *abg, int beg_index, int end_index) {
-    int min_index, i, j, node_id, in_id;
+    int min_index, in_index, i, j, node_id, in_id;
 
     while (1) {
-        min_index = INT32_MAX;
+        min_index = beg_index;
         for (i = beg_index; i <= end_index; ++i) {
             node_id = abg->index_to_node_id[i];
             for (j = 0; j < abg->node[node_id].in_edge_n; ++j) {
                 in_id = abg->node[node_id].in_id[j];
-                min_index = MIN_OF_TWO(min_index, abg->node_id_to_index[in_id]);
+                in_index = abg->node_id_to_index[in_id];
+                min_index = MIN_OF_TWO(min_index, in_index);
             }
         }
         if (is_full_upstream_subgraph(abg, min_index, beg_index)) {
@@ -1131,28 +1170,32 @@ int is_full_downstream_subgraph(abpoa_graph_t *abg, int up_index, int down_index
 }
 
 int abpoa_downstream_index(abpoa_graph_t *abg, int beg_index, int end_index) {
-    int max_index, i, j, node_id, out_id;
+    int max_index, out_index, i, j, node_id, out_id;
 
     while (1) {
-        max_index = -1;
+        max_index = end_index;
         for (i = beg_index; i <= end_index; ++i) {
             node_id = abg->index_to_node_id[i];
             for (j = 0; j < abg->node[node_id].out_edge_n; ++j) {
                 out_id = abg->node[node_id].out_id[j];
-                max_index = MAX_OF_TWO(max_index, abg->node_id_to_index[out_id]);
+                out_index = abg->node_id_to_index[out_id];
+                max_index = MAX_OF_TWO(max_index, out_index);
             }
         }
         if (is_full_upstream_subgraph(abg, end_index, max_index)) {
             return max_index;
         } else {
-            beg_index = end_index; 
+            beg_index = end_index;
             end_index = max_index;
         }
-    }}
+    }
+}
+
 //   exc_beg | inc_beg ... inc_end | exc_end
-void abpoa_subgraph_nodes(abpoa_t *ab, int inc_beg, int inc_end, int *exc_beg, int *exc_end) {
+void abpoa_subgraph_nodes(abpoa_t *ab, abpoa_para_t *abpt, int exc_beg0, int exc_end0, int *exc_beg, int *exc_end) {
     abpoa_graph_t *abg = ab->abg;
-    int inc_beg_index = abg->node_id_to_index[inc_beg], inc_end_index = abg->node_id_to_index[inc_end];
+    if (ab->abg->is_topological_sorted == 0) abpoa_topological_sort(abg, abpt);
+    int inc_beg_index = abg->node_id_to_index[exc_beg0], inc_end_index = abg->node_id_to_index[exc_end0];
 
     int exc_beg_index = abpoa_upstream_index(abg, inc_beg_index, inc_end_index);
     int exc_end_index = abpoa_downstream_index(abg, inc_beg_index, inc_end_index);
@@ -1170,19 +1213,15 @@ void abpoa_subgraph_nodes(abpoa_t *ab, int inc_beg, int inc_end, int *exc_beg, i
 // 4. Deletion: nothing
 // 5. Clipping: add node
 // 6. For all first/last node, link to virtual start/end node
-//33S:32
-//1M:74,33
-//26I:59
 
-// for seq-to-graph alignment
-//     generate dot file
-int abpoa_add_subgraph_alignment(abpoa_t *ab, abpoa_para_t *abpt, int beg_node_id, int end_node_id, uint8_t *seq, int seq_l, abpoa_res_t res, int read_id, int tot_read_n) {
+// inc_both_ends: set as 1 to add weight for edge between beg_node_id/end_node_id and internal node
+int abpoa_add_subgraph_alignment(abpoa_t *ab, abpoa_para_t *abpt, int beg_node_id, int end_node_id, uint8_t *seq, int seq_l, int *qpos_to_node_id, abpoa_res_t res, int read_id, int tot_read_n, int inc_both_ends) {
     abpoa_graph_t *abg = ab->abg;
     int n_cigar = res.n_cigar; abpoa_cigar_t *abpoa_cigar = res.graph_cigar;
     int read_ids_n = 1 + ((tot_read_n-1) >> 6);
-    uint8_t add_read_id = abpt->use_read_ids, add;
+    uint8_t add_read_id = abpt->use_read_ids, add, w=1;
     if (abg->node_n == 2) { // empty graph
-        abpoa_add_graph_sequence(abg, seq, seq_l, 0, seq_l, add_read_id, read_id, read_ids_n);
+        abpoa_add_graph_sequence(abg, seq, seq_l, qpos_to_node_id, 0, seq_l, add_read_id, read_id, read_ids_n);
         return 0;
     } else {
         if (abg->node_n < 2) {
@@ -1193,65 +1232,59 @@ int abpoa_add_subgraph_alignment(abpoa_t *ab, abpoa_para_t *abpt, int beg_node_i
         }
     }
     // normal graph, normal graph_cigar
-    int i, j; int op, len, node_id, query_id, last_new = 0, last_id = beg_node_id, new_id, aligned_id;
-    int w; //
-    if (res.is_rc) { // reverse complement
-        uint8_t tmp1, tmp2;
-        for (i = 0; i < (seq_l+1)/2; ++i) {
-            tmp1 = seq[i], tmp2 = seq[seq_l-i-1];
-            if (tmp1 < 4) seq[seq_l-i-1] = 3 - tmp1;
-            else seq[seq_l-i-1] = 4;
-            if (tmp2 < 4) seq[i] = 3 - tmp2;
-            else seq[i] = 4;
-        }
-    }
+    int i, j; int op, len;
+    int node_id, query_id=-1, last_new = 0, last_id = beg_node_id, new_id, aligned_id;
+
     for (i = 0; i < n_cigar; ++i) {
         op = abpoa_cigar[i] & 0xf;
         if (op == ABPOA_CMATCH) {
             node_id = (abpoa_cigar[i] >> 34) & 0x3fffffff;
-            query_id = (abpoa_cigar[i] >> 4) & 0x3fffffff;
+            query_id++; // = (abpoa_cigar[i] >> 4) & 0x3fffffff;
             if (abg->node[node_id].base != seq[query_id]) { // mismatch
                 // check if query base is identical to node_id's aligned node
-                if ((aligned_id = abpoa_get_aligned_id(abg, node_id, seq[query_id])) >= 0) {
-                    if (last_id == beg_node_id && beg_node_id != ABPOA_SRC_NODE_ID) w = 0, add = 0; else w = 1, add = 1;
+                if ((aligned_id = abpoa_get_aligned_id(abg, node_id, seq[query_id])) != -1) {
+                    if (last_id != beg_node_id || inc_both_ends) add = 1; else add = 0;
                     abpoa_add_graph_edge(abg, last_id, aligned_id, 1-last_new, w, add_read_id&add, read_id, read_ids_n);
                     last_id = aligned_id; last_new = 0;
                 } else {
                     new_id = abpoa_add_graph_node(abg, seq[query_id]);
-                    if (last_id == beg_node_id && beg_node_id != ABPOA_SRC_NODE_ID) w = 0, add = 0; else w = 1, add = 1;
+                    if (last_id != beg_node_id || inc_both_ends) add = 1; else add = 0;
                     abpoa_add_graph_edge(abg, last_id, new_id, 0, w, add_read_id&add, read_id, read_ids_n);
                     last_id = new_id; last_new = 1;
                     // add new_id to node_id's aligned node
                     abpoa_add_graph_aligned_node(abg, node_id, new_id);
                 }
             } else { // match
-                if (last_id == beg_node_id && beg_node_id != ABPOA_SRC_NODE_ID) w = 0, add = 0; else w = 1, add = 1;
+                if (last_id != beg_node_id || inc_both_ends) add = 1; else add = 0;
                 abpoa_add_graph_edge(abg, last_id, node_id, 1-last_new, w, add_read_id&add, read_id, read_ids_n);
                 last_id = node_id; last_new = 0;
             }
+            if (qpos_to_node_id) qpos_to_node_id[query_id] = last_id;
         } else if (op == ABPOA_CINS || op == ABPOA_CSOFT_CLIP || op == ABPOA_CHARD_CLIP) {
-            query_id = (abpoa_cigar[i] >> 34) & 0x3fffffff;
             len = (abpoa_cigar[i] >> 4) & 0x3fffffff;
+            query_id+=len; // = (abpoa_cigar[i] >> 34) & 0x3fffffff;
             for (j = len-1; j >= 0; --j) { // XXX use dynamic id, instead of static query_id
                 new_id = abpoa_add_graph_node(abg, seq[query_id-j]);
-                if (last_id == beg_node_id && beg_node_id != ABPOA_SRC_NODE_ID) w = 0, add = 0; else w = 1, add = 1;
+                if (last_id != beg_node_id || inc_both_ends) add = 1; else add = 0;
                 abpoa_add_graph_edge(abg, last_id, new_id, 0, w, add_read_id&add, read_id, read_ids_n);
                 last_id = new_id; last_new = 1;
+                if (qpos_to_node_id) qpos_to_node_id[query_id-j] = last_id;
             }
         } else if (op == ABPOA_CDEL) {
             // nothing;
             continue;
         }
     } 
-    if (last_id == beg_node_id && beg_node_id != ABPOA_SRC_NODE_ID) w = 0, add = 0; else w = 1, add = 1;
-    abpoa_add_graph_edge(abg, last_id, end_node_id, 1-last_new, w, add_read_id&add, read_id, read_ids_n);
+    // if (inc_both_ends) add = 1; else add = 0; XXX end_node_id is always excluded when adding weight
+    // abpoa_add_graph_edge(abg, last_id, end_node_id, 1-last_new, w, add_read_id&add, read_id, read_ids_n);
+    abpoa_add_graph_edge(abg, last_id, end_node_id, 1-last_new, w, add_read_id, read_id, read_ids_n);
     abg->is_called_cons = abg->is_topological_sorted = 0;
     // abpoa_topological_sort(abg, abpt);
     return 0;
 }
 
-int abpoa_add_graph_alignment(abpoa_t *ab, abpoa_para_t *abpt, uint8_t *seq, int seq_l, abpoa_res_t res, int read_id, int tot_read_n) {
-    return abpoa_add_subgraph_alignment(ab, abpt, ABPOA_SRC_NODE_ID, ABPOA_SINK_NODE_ID, seq, seq_l, res, read_id, tot_read_n);
+int abpoa_add_graph_alignment(abpoa_t *ab, abpoa_para_t *abpt, uint8_t *seq, int seq_l, int *qpos_to_node_id, abpoa_res_t res, int read_id, int tot_read_n, int inc_both_ends) {
+    return abpoa_add_subgraph_alignment(ab, abpt, ABPOA_SRC_NODE_ID, ABPOA_SINK_NODE_ID, seq, seq_l, qpos_to_node_id, res, read_id, tot_read_n, inc_both_ends);
 }
 
 // reset allocated memery everytime init the graph
